/**
 * Author:  John Idarraga <idarraga@cern.ch> , 2011
 *  An example of how to use the Clustering results
 *  for further processing.
 */

#ifndef ShallowAngleFEIX_cpp
#define ShallowAngleFEIX_cpp

#include "ShallowAngleFEIX.h"
#include "MAFTools.h"

using namespace MSG;

ClassImp(ShallowAngleFEIX)

ShallowAngleFEIX::ShallowAngleFEIX()
:
m_pixSizeX ( 0.40 ), // FEI3 [mm]
m_pixSizeY ( 0.05 ), // FEI3 [mm]
m_minXWidth ( 6 ),
m_guardDistanceX ( 1 ),
m_dRayBalSearch ( 1.4 ),
m_guardDistanceY ( 5 ),
m_minNPixels ( 5 )
{

}

void ShallowAngleFEIX::Init(){

	// You will get an ntuple file containing a TTree with the name of this
	//  Algorithm.  The branches registered through getMyTree() get registered
	//  in that tree so you can fill them each time Execute() gets called.
	getMyTree()->Branch("distanceToCenter", &m_distanceToCenter);
	getMyTree()->Branch("clusterSize", &m_clusterSize, "clusterSize/I");
	getMyTree()->Branch("mipLength", &m_mipLength);
	getMyTree()->Branch("mipLength_X", &m_mipLength_X);
	getMyTree()->Branch("totPerSegment", &m_totPerSegment);
	getMyTree()->Branch("chargeWeights", &m_chargeWeights);
	getMyTree()->Branch("clusterHeight", &m_clusterHeight, "clusterHeight/I");

	// A configuration value that can be tuned from the Viewer
	RegisterConfigurationValue(&m_minNPixels, "minNPixels");
	RegisterConfigurationValue(&m_minXWidth, "minXWidth");
	RegisterConfigurationValue(&m_guardDistanceX, "guardDistanceX");
	RegisterConfigurationValue(&m_guardDistanceY, "guardDistanceY");
	RegisterConfigurationValue(&m_pixSizeX, "pixelSizeX");
	RegisterConfigurationValue(&m_pixSizeY, "pixelSizeY");
	RegisterConfigurationValue(&m_dRayBalSearch, "dRayBalSearch");

}

void ShallowAngleFEIX::Execute(){

	// Ask the store gate if the previous algorithm (BlobsFinder --> reponsible for clustering)
	//  sent any objects to the StoreGate.
	Int_t lastObject = GetNumberOfObjectsWithAuthor("BlobsFinder");
	if(lastObject == 0)
		return;

	// If so, get the pointer to the last object.  BlobsFinder, as it comes out of the box, sends
	//  a single object containing all clusters.
	m_aB = (AllBlobsContainer *) GetObjectFromAuthor("BlobsFinder", lastObject-1);

	// AllBlobsContainer is a box full of Blobs(Clusters). Now we can iterate over all clusters inside.
	vector<blob> blobsVector = m_aB->GetBlobsVector();
	Log << MSG::INFO << "Number of blobs from clustering = " << (Int_t) blobsVector.size() << endreq;
	vector<blob>::iterator blobsItr = blobsVector.begin(); //allBlobs.begin();


	for( ; blobsItr != blobsVector.end() ; blobsItr++)
	{

		// Limit all this to clusters with a minimum size.
		// Note that m_minNPixels can be configured through the Viewer
		//  so you can reprocess and check results online.
		if((*blobsItr).bP.nPixels < m_minNPixels)
			continue;

		// Look for mips
		blobtype bT = (*blobsItr).GetBlobType();

		if(bT == _MIP) { // only Mips

			// info about the mip to store
			m_clusterSize = (*blobsItr).GetBlobProperties().nPixels;

			// Here we obtain two pairs with the pixels at the extreme ends of the Mip
			vector< pair<Float_t, Float_t> > limitPixels = FindLimitPixels(*blobsItr);

			// Make sure the mip starts and ends inside the pad at a certain minimum
			//  distance from the borders
			if ( MAFTools::BlobAtADistanceFromEdges(*blobsItr, limitPixels, GetWidth(), GetHeight(),
					m_guardDistanceX, m_guardDistanceY) ) {
				continue;
			}

			// If the whole mip is in the same column, don't process it.
			// Problems with reading of too many pixels activated in the
			//  same column in FE-I3.  FE-I4 too ?
			if (limitPixels[__TAIL_INDX].first - limitPixels[__HEAD_INDX].first == 0) {
				continue;
			}

			// Check if the mip is long enough
			// working this out in units of pixels
			if ( (*blobsItr).GetBlobProperties().width_x < m_minXWidth )
				continue;

			// OK, if the mip is fine, collect a few stats.
			//  This is distance, as opposed to that calculated for divOnTrack, has to have
			//  real units.
			Double_t totalLength = MAFTools::CalcDistance(limitPixels[__HEAD_INDX].first * m_pixSizeX,
					limitPixels[__HEAD_INDX].second * m_pixSizeY,
					limitPixels[__TAIL_INDX].first * m_pixSizeX,
					limitPixels[__TAIL_INDX].second * m_pixSizeY);
			m_mipLength.push_back( totalLength ); // mm
			m_mipLength_X.push_back( TMath::Abs ( limitPixels[__HEAD_INDX].first * m_pixSizeX
					- limitPixels[__TAIL_INDX].first * m_pixSizeX ) );
			m_totalMIPsLength += totalLength;
			Log << MSG::DEBUG << "Integrated mips length = " << m_totalMIPsLength << " [mm]" << endreq;

			// Fech the linear regression information for this mip
			Float_t slope = (*blobsItr).GetBlobProperties().fitSlope;
			Float_t cut = (*blobsItr).GetBlobProperties().fitCut;

			// Draw the linear regression and the line which passes through the limits
			MAFTools::DrawLine(this, limitPixels[__HEAD_INDX].first, limitPixels[__HEAD_INDX].second,
					limitPixels[__TAIL_INDX].first, limitPixels[__TAIL_INDX].second, 2, 2, kBlack);
			MAFTools::DrawLine(this, limitPixels[__HEAD_INDX].first, limitPixels[__HEAD_INDX].first * slope + cut,
					limitPixels[__TAIL_INDX].first, limitPixels[__TAIL_INDX].first * slope + cut, 2, 2, kRed);


			Log << MSG::INFO << "Mip angle = " <<  (*blobsItr).GetBlobProperties().rotAngle << endreq;
			if( TMath::Abs( (*blobsItr).GetBlobProperties().rotAngle ) > TMath::Pi()/3.)
				continue;

			// Find the limit pixels to add-up the charge per column.
			//  It turns out that the limit pixels found above are not exactly
			//  what I need here.  Often a delta ray comes out of the box marked
			//  by these limits and it could remain not included.
			// I will use the limits given by the box.

			Float_t midx = ((Float_t)(*blobsItr).GetBlobProperties().width_x) /2.;
			Float_t midy = ((Float_t)(*blobsItr).GetBlobProperties().width_y) /2.;
			int initx = TMath::FloorNint((*blobsItr).GetBlobProperties().geoCenter_x - midx);
			int endx = TMath::FloorNint((*blobsItr).GetBlobProperties().geoCenter_x + midx - 1);
			int inity = TMath::FloorNint((*blobsItr).GetBlobProperties().geoCenter_y - midy);
			int endy = TMath::FloorNint((*blobsItr).GetBlobProperties().geoCenter_y + midy - 1);

			Log << MSG::INFO << "Searching in box limited by x:" << initx << " , " << endx << endreq;
			Log << MSG::INFO << "                            y:" << inity << " , " << endy << endreq;

			m_clusterHeight = (*blobsItr).GetBlobProperties().width_y;

			int cntr = 0;
			for(int x = initx ; x <= endx ; x++){

				m_totPerSegment.push_back(0); // add segment initialized at 0
				m_chargeWeights.push_back(0); // add segment

				for (int y = inity ; y <= endy ; y++) {

					Log << MSG::DEBUG << "  --> " << x << " , " << y << " : " << GetMatrixElement(x,y) << endreq;
					m_totPerSegment[cntr] += GetMatrixElement(x,y);
					m_chargeWeights[cntr] += GetMatrixElement(x,y); // normalized later

				}
				cntr++;
			}

			// normalize m_chargeWeights to total charge
			vector<Float_t>::iterator itrCW = m_chargeWeights.begin();
			for ( ; itrCW != m_chargeWeights.end() ; itrCW++ ) {
				(*itrCW) /= (*blobsItr).GetBlobProperties().totalCharge;
			}

			// DEBUG output.  Charge fraction and TOT

			int segCntr = 0;
			if(Log.OutputLevel == MSG::DEBUG) {

				vector<Float_t>::iterator itrW = m_chargeWeights.begin();
				vector<Int_t>::iterator itrT = m_totPerSegment.begin();

				for( ; itrW != m_chargeWeights.end() ; ) {

					Log << MSG::DEBUG << "[" << segCntr << "] "
							<< "(TOT , fraction) : " <<
							*itrT << " , " << (*itrW)*100 << "\%"
							<< endreq;

					itrW++;
					itrT++;
					segCntr++;
				}

			}

			// Perform DeltaRay identification by excessive raise in the signal along the track
			Int_t nDeltaR = 0;
			Int_t nDeltaSoft = 0;
			GuessNumberOfDeltaRaysFromUnbalance(nDeltaR, nDeltaSoft);


		}

		// there might be more than 1 good mip in a frame
		// Fill and clean up here

		// Fill the output tree of this algorithm
		getMyTree()->Fill();

		// WARNING ! don't forget to clean up your variables for the next TTree::Fill call
		m_distanceToCenter.clear();
		m_totPerSegment.clear();
		m_chargeWeights.clear();
		m_mipLength.clear();
		m_mipLength_X.clear();

	}


}

Int_t ShallowAngleFEIX::GuessNumberOfDeltaRaysFromUnbalance(Int_t & nDR, Int_t & nDRSoft){

	vector<Float_t>::iterator itr = m_chargeWeights.begin();
	vector<Float_t>::iterator itrC = m_chargeWeights.begin();
	Int_t nDiv = (Int_t) m_chargeWeights.size();
	vector<Int_t> nNoticeableUnbalances( nDiv );  // initialize at 0

	Float_t unbalance = 1.;
	Int_t cntr = 0, cntrE = 0;
	Int_t nExplicit = 0, nSoft = 0;

	Log << MSG::INFO << "Starting delta ray search " << endreq;

	for ( ; itr != m_chargeWeights.end() ; itr++ ) {

		cntrE = 0;

		for ( itrC = m_chargeWeights.begin() ; itrC != m_chargeWeights.end() ; itrC++ ) {

			// unbalance between the two fragments
			unbalance = (*itr)/(*itrC);
			if ( unbalance > m_dRayBalSearch ) {
				Log << MSG::DEBUG << "section " << cntr << " much hotter than " << cntrE << endreq;
				nNoticeableUnbalances[cntr]++;
			}

			cntrE++;
		}

		cntr++;
	}

	// based on nNoticeableUnbalances and m_nDivisions determine if there is a
	// delta ray or not

	vector<Int_t>::iterator itrU = nNoticeableUnbalances.begin();  // initialize at 0

	for( ; itrU != nNoticeableUnbalances.end() ; itrU++){

		if ( (*itrU) == nDiv - 1 ) { // an explicit delta ray.  Un-balance to all other segments
			nExplicit++;
		} else if ( (*itrU) == nDiv - 2 ) { // a soft.  Unbalance to all-1 other segments
			nSoft++;
		}

	}

	Log << MSG::INFO << "Explicit : " << nExplicit << " , Soft : " << nSoft << endreq;

	nDR = nExplicit;
	nDRSoft = nSoft;

	return nExplicit;
}

vector<pair<Float_t, Float_t> > ShallowAngleFEIX::FindLimitPixels(blob theBlob) {

	// result vector
	vector<pair<Float_t, Float_t> > limits;
	limits.push_back( make_pair(0, 0) );
	limits.push_back( make_pair(0, 0) );

	set< pair<Int_t, Int_t> > contents = theBlob.GetContentSet();

	set< pair<Int_t, Int_t> >::iterator contItrA = contents.begin();
	set< pair<Int_t, Int_t> >::iterator contItrB = contents.begin();

	double max = 0., dist = 0.;

	for( ; contItrA != contents.end() ; contItrA++) {

		for(contItrB = contItrA ; contItrB != contents.end() ; contItrB++) {

			dist = MAFTools::CalcDistance( *contItrA, *contItrB );

			if( dist > max ) {
				max = dist;
				limits[0] = *contItrA;
				limits[1] = *contItrB;
			}

		}

	}

	return limits;
}

void ShallowAngleFEIX::Finalize() {

	Log << MSG::INFO << "Finalize function !" << endreq;

}

#endif
